#!/usr/bin/env python

"""
Acorn Electron ULA simulation.

Copyright (C) 2011, 2012, 2013, 2014, 2016 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from array import array
from itertools import repeat

LINES_PER_ROW = 8       # the number of pixel lines per character row
MAX_HEIGHT = 256        # the height of the screen in pixels
MAX_WIDTH = 640         # the width of the screen in pixels

MAX_CSYNC = 2           # the scanline during which vsync ends
MIN_PIXELLINE = 28      # the first scanline involving pixel generation
MAX_SCANLINE = 312      # the number of scanlines in each frame

MAX_PIXELLINE = MIN_PIXELLINE + MAX_HEIGHT

MAX_HSYNC = 64          # the number of cycles in each hsync period
MIN_PIXELPOS = 256      # the first cycle involving pixel generation
MAX_SCANPOS = 1024      # the number of cycles in each scanline

MAX_PIXELPOS = MIN_PIXELPOS + MAX_WIDTH

SCREEN_LIMIT = 0x8000   # the first address after the screen memory
MAX_MEMORY = 0x10000    # the number of addressable memory locations
MAX_RAM = 0x10000       # the number of addressable RAM locations (64Kb in each IC)
BLANK = (0, 0, 0)

def update(ula):

    """
    Update the 'ula' for one frame. Return the resulting screen.
    """

    video = ula.video

    i = 0
    limit = MAX_SCANLINE * MAX_SCANPOS
    while i < limit:
        ula.posedge()
        video.update()
        ula.negedge()
        i += 1

    return video.screen

class Video:

    """
    A class representing the video circuitry.
    """

    def __init__(self):
        self.screen = array("B", repeat(0, MAX_WIDTH * 3 * MAX_HEIGHT))
        self.colour = BLANK
        self.csync = 1
        self.hs = 1
        self.x = 0
        self.y = 0

    def set_csync(self, value):
        if self.csync and not value:
            self.y = 0
            self.pos = 0
        self.csync = value

    def set_hs(self, value):
        if self.hs and not value:
            self.x = 0
            self.y += 1
        self.hs = value

    def update(self):
        if MIN_PIXELLINE <= self.y < MAX_PIXELLINE:
            if MIN_PIXELPOS + 8 <= self.x < MAX_PIXELPOS + 8:
                self.screen[self.pos] = self.colour[0]; self.pos += 1
                self.screen[self.pos] = self.colour[1]; self.pos += 1
                self.screen[self.pos] = self.colour[2]; self.pos += 1
            self.x += 1

class RAM:

    """
    A class representing the RAM circuits (IC4 to IC7). Each circuit
    traditionally holds 64 kilobits, with each access obtaining 1 bit from each
    IC, and thus two accesses being required to obtain a whole byte. Here, we
    model the circuits with a list of 65536 half-bytes with each bit in a
    half-byte representing a bit stored on a separate IC.
    """

    def __init__(self):

        "Initialise the RAM circuits."

        self.memory = [0] * MAX_RAM
        self.row_address = 0
        self.column_address = 0
        self.data = 0
        self.read_not_write = 1

    def row_select(self, address):

        "The operation of asserting a row 'address' via RA0...RA7."

        self.row_address = address

    def row_deselect(self):
        pass

    def column_select(self, address):

        "The operation of asserting a column 'address' via RA0...RA7."

        self.column_address = address

        # Read the data.

        if self.read_not_write:
            self.data = self.memory[self.row_address << 8 | self.column_address]
        else:
            self.memory[self.row_address << 8 | self.column_address] = self.data

    def column_deselect(self):
        pass

    def read_select(self):
        self.read_not_write = 1

    def write_select(self):
        self.read_not_write = 0

    # Convenience methods.

    def fill(self, start, end, value):
        for i in xrange(start, end):
            self.memory[i << 1] = value >> 4
            self.memory[i << 1 | 0x1] = value & 0xf

class CPU:

    "A CPU abstraction."

    def __init__(self):
        self.address = 0x1000
        self.data = 0
        self.read_not_write = 1

class ULA:

    """
    A class providing the ULA functionality. Instances of this class refer to
    the system memory, maintain internal state (such as information about the
    current screen mode), and provide outputs (such as the current pixel
    colour).
    """

    modes = [
        (640, 1, 32), (320, 2, 32), (160, 4, 32),   # (width, depth, rows)
        (640, 1, 25), (320, 1, 32), (160, 2, 32),
        (320, 1, 25)
        ]

    def __init__(self, cpu, ram, video):

        "Initialise the ULA with the given 'cpu', 'ram' and 'video' instances."

        self.cpu = cpu
        self.ram = ram
        self.video = video
        self.set_mode(6)
        self.palette = map(get_physical_colour, range(0, 8) * 2)

        self.reset()

    def reset(self):

        "Reset the ULA."

        # General state.

        self.nmi = 0            # no NMI asserted initially
        self.irq_vsync = 0      # no IRQ asserted initially
        self.cpu_clock = 0      # drive the CPU clock low

        # Communication.

        self.ram_address = 0    # address given to the RAM via RA0...RA7
        self.data = 0           # data read from the RAM via RAM0...RAM3
        self.cpu_address = 0    # address selected by the CPU via A0...A15
        self.cpu_read = 0       # data read/write by the CPU selected using R/W

        # Internal state.

        self.have_pixels = 0    # whether pixel data has been read
        self.pdata = 0          # decoded RAM data for pixel output
        self.cycle = 1          # 8-state counter within each 2MHz period
        self.pcycle = 0         # 8/4/2-state pixel output counter

        self.next_frame()

    def set_mode(self, mode):

        """
        For the given 'mode', initialise the...

          * width in pixels
          * colour depth in bits per pixel
          * number of character rows
          * character row size in bytes
          * screen size in bytes
          * default screen start address
          * horizontal pixel scaling factor
          * row height in pixels
          * display height in pixels

        The ULA should be reset after a mode switch in order to cleanly display
        a full screen.
        """

        self.width, self.depth, rows = self.modes[mode]

        columns = (self.width * self.depth) / 8                             # bits read -> bytes read
        self.access_frequency = 80 / columns                                # cycle frequency for reading bytes
        row_size = columns * LINES_PER_ROW

        # Memory access configuration.
        # Note the limitation on positioning the screen start.

        screen_size = row_size * rows
        self.screen_start = (SCREEN_LIMIT - screen_size) & 0xff00
        self.screen_size = SCREEN_LIMIT - self.screen_start

        # Scanline configuration.

        self.xscale = MAX_WIDTH / self.width                                # pixel width in display pixels
        self.row_height = MAX_HEIGHT / rows                                 # row height in display pixels
        self.display_height = rows * self.row_height                        # display height in pixels

    def vsync(self, value=0):

        "Signal the start of a frame."

        self.csync = value
        self.video.set_csync(value)

    def hsync(self, value=0):

        "Signal the end of a scanline."

        self.hs = value
        self.video.set_hs(value)

    def next_frame(self):

        "Signal the start of a frame."

        self.line_start = self.pixel_address = self.screen_start
        self.line = self.line_start % LINES_PER_ROW
        self.y = 0
        self.x = 0

    def next_horizontal(self):

        "Visit the next horizontal position."

        self.pixel_address += LINES_PER_ROW
        self.wrap_address()

    def next_vertical(self):

        "Reset horizontal state within the active region of the frame."

        self.y += 1
        self.x = 0

        if self.inside_frame():
            self.line += 1

            # At the end of a row...

            if self.line == self.row_height:

                # After the end of the last line in a row, the address should already
                # have been positioned on the last line of the next column.

                self.pixel_address -= LINES_PER_ROW - 1
                self.wrap_address()
                self.line = 0

                # Record the position of the start of the pixel row.

                self.line_start = self.pixel_address

            # Before any spacing between character rows...

            elif self.in_line():

                # If not on a row boundary, move to the next line. Here, the address
                # needs bringing back to the previous character row.

                self.pixel_address = self.line_start + 1
                self.wrap_address()

                # Record the position of the start of the pixel row.

                self.line_start = self.pixel_address

    def access_cycle(self): return (self.x / 8) % self.access_frequency == 0 
    def would_access_ram(self): return self.access_cycle() and self.read_pixels() and self.in_line()
    def access_ram(self): return not self.nmi and self.would_access_ram()
    def in_line(self): return self.line < LINES_PER_ROW
    def in_frame(self): return MIN_PIXELLINE <= self.y < (MIN_PIXELLINE + self.display_height)
    def inside_frame(self): return MIN_PIXELLINE < self.y < (MIN_PIXELLINE + self.display_height)
    def read_pixels(self): return MIN_PIXELPOS <= self.x < MAX_PIXELPOS and self.in_frame()
    def write_pixels(self): return self.pcycle != 0
    def next_pixel(self): return self.xscale == 1 or (self.xscale == 2 and self.cycle & 0b01010101) or (self.xscale == 4 and self.cycle & 0b00010001)

    def posedge(self):

        "Update the state of the ULA for each clock cycle."

        self.posedge_video()
        self.posedge_ram()
        self.posedge_pixel()

    def posedge_video(self):

        "Video signalling."

        # Detect the end of the scanline.

        if self.x == MAX_SCANPOS:
            self.next_vertical()

        # Detect the end of the frame.

        if self.y == MAX_SCANLINE:
            self.next_frame()

        # Detect any sync conditions.

        if self.x == 0:
            self.hsync()
            if self.y == 0:
                self.vsync()
                self.irq_vsync = 0
            elif self.y == MAX_PIXELLINE:
                self.irq_vsync = 1

        # Detect the end of hsync.

        elif self.x == MAX_HSYNC:
            self.hsync(1)

        # Detect the end of vsync.

        elif self.y == MAX_CSYNC and self.x == MAX_SCANPOS / 2:
            self.vsync(1)

    def posedge_ram(self):

        "RAM signalling."

        # Clock management.

        # Reset addresses.

        if self.cycle == 1:
            self.ram.column_deselect()
            self.ram.row_deselect()

            # Read the CPU address, if appropriate.

            if not self.access_ram():
                self.cpu_update_clock()

        # Set row address (for ULA access only).

        elif self.cycle == 2:

            # Either assert a required address or propagate the CPU address.

            if self.access_ram():
                self.init_row_address(self.pixel_address)
            else:
                self.init_row_address(self.cpu_address)

        # Latch row address, set column address (for ULA access only).

        elif self.cycle == 4:

            # Select an address needed by the ULA or CPU.

            self.ram.row_select(self.ram_address)

            # Either assert a required address or propagate the CPU address.

            if self.access_ram():
                self.init_column_address(self.pixel_address, 0)
            else:
                self.init_column_address(self.cpu_address, 0)

        # Latch column address.

        elif self.cycle == 8:

            # Select an address needed by the ULA or CPU.

            self.ram.column_select(self.ram_address)

            # Assert the RAM write enable if appropriate.

            if self.access_ram():
                self.ram.read_select()
            else:
                self.cpu_transfer_select()

        # Cycle handled in negedge.

        # Set column address (for ULA access only).

        elif self.cycle == 32:
            self.ram.column_deselect()

            # Either assert a required address or propagate the CPU address.

            if self.access_ram():
                self.init_column_address(self.pixel_address, 1)
            else:
                self.init_column_address(self.cpu_address, 1)

        # Latch column address.

        elif self.cycle == 64:

            # Select an address needed by the ULA or CPU.

            self.ram.column_select(self.ram_address)

        # Read 4 bits (for ULA access only).

        elif self.cycle == 128:

            # Advance to the next column even if an NMI is asserted.

            if self.would_access_ram():
                self.next_horizontal()

    def posedge_pixel(self):

        "Pixel production."

        # For pixels within the frame, obtain and output the value.

        if self.write_pixels():
            self.output_colour_value()

            # Scale pixels horizontally, only accessing the next pixel value
            # after the required number of scan positions.

            if self.next_pixel():
                self.next_pixel_value()

        # Detect spacing between character rows.

        else:
            self.video.colour = BLANK

    def negedge(self):

        """
        Update the state of the device.

        Cycles handled: * _ _ _ * _ _ *
        """

        # Clock management.

        # Initialise the pixel buffer if appropriate. Output starts after
        # this cycle.

        if self.cycle == 1 and self.have_pixels:
            self.pdata = decode(self.data, self.depth)
            self.pcycle = 1
            self.have_pixels = 0

        # Read 4 bits (for ULA access only).

        elif self.cycle == 16:

            # Either read from a required address or transfer CPU data.

            if self.access_ram():
                self.data = self.ram.data << 4
            else:
                self.cpu_transfer_high()

        # Read 4 bits (for ULA access only).

        elif self.cycle == 128:

            # Either read from a required address or transfer CPU data.

            if self.access_ram():
                self.data = self.data | self.ram.data
                self.have_pixels = 1
            else:
                self.cpu_transfer_low()

        # Start a new cycle.

        self.cycle = rotate(self.cycle, 1)
        self.x += 1

    def output_colour_value(self):

        """
        Output the colour value for the current pixel by translating memory
        content for the current mode.
        """

        value = value_of_bits(self.pdata, self.depth)
        self.video.colour = self.palette[value]

    def next_pixel_value(self):
        self.pdata = rotate(self.pdata, self.depth)
        self.pcycle = rotate(self.pcycle, self.depth, zero=True)

    def wrap_address(self):
        if self.pixel_address >= SCREEN_LIMIT:
            self.pixel_address -= self.screen_size

    def init_row_address(self, address):
        self.ram_address = (address & 0xff80) >> 7

    def init_column_address(self, address, offset):
        self.ram_address = (address & 0x7f) << 1 | offset

    def cpu_transfer_high(self):
        if self.cpu_read:
            self.cpu_data = self.ram.data << 4
        else:
            self.ram.data = self.cpu_data >> 4

    def cpu_transfer_low(self):
        if self.cpu_read:
            self.cpu_data = self.data | self.ram.data
        else:
            self.ram.data = self.cpu_data & 0b00001111

    def cpu_read_address(self):
        self.cpu_address = self.cpu.address

    def cpu_transfer_data(self):
        if self.cpu_read:
            self.cpu.data = self.cpu_data
        else:
            self.cpu_data = self.cpu.data

    def cpu_update_clock(self):
        self.cpu_clock = not self.cpu_clock
        if self.cpu_clock:
            self.cpu_transfer_data()
        else:
            self.cpu_read_address()

    def cpu_transfer_select(self):
        self.cpu_read = self.cpu.read_not_write
        if self.cpu_read:
            self.ram.read_select()
        else:
            self.ram.write_select()

def rotate(value, depth, width=8, zero=False):

    """
    Return 'value' rotated left by the number of bits given by 'depth', doing so
    within a value 'width' given in bits. If 'zero' is true, rotate zero bits
    into the lower bits when rotating.
    """

    field = width - depth
    top = value >> field
    mask = 2 ** (width - depth) - 1
    rest = value & mask
    return (rest << depth) | (not zero and top or 0)

def value_of_bits(value, depth):

    """
    Convert the upper bits of 'value' to a result, using 'depth' to indicate the
    number of bits involved.
    """

    return value >> (8 - depth)

def get_physical_colour(value):

    """
    Return the physical colour as an RGB triple for the given 'value'.
    """

    return value & 1, value >> 1 & 1, value >> 2 & 1

def decode(value, depth):

    """
    Decode the given byte 'value' according to the 'depth' in bits per pixel,
    returning a sequence of pixel values.
    """

    if depth == 1:
        return value
    elif depth == 2:
        return ((value & 128) | ((value & 8) << 3) | ((value & 64) >> 1) | ((value & 4) << 2) |
                ((value & 32) >> 2) | ((value & 2) << 1) | ((value & 16) >> 3) | (value & 1))
    elif depth == 4:
        return ((value & 128) | ((value & 32) << 1) | ((value & 8) << 2) | ((value & 2) << 3) |
                ((value & 64) >> 3) | ((value & 16) >> 2) | ((value & 4) >> 1) | (value & 1))
    else:
        raise ValueError, "Only depths of 1, 2 and 4 are supported, not %d." % depth

# Convenience functions.

def encode(values, depth):

    """
    Encode the given 'values' according to the 'depth' in bits per pixel,
    returning a byte value for the pixels.
    """

    result = 0

    if depth == 1:
        for value in values:
            result = result << 1 | (value & 1)
    elif depth == 2:
        for value in values:
            result = result << 1 | (value & 2) << 3 | (value & 1)
    elif depth == 4:
        for value in values:
            result = result << 1 | (value & 8) << 3 | (value & 4) << 2 | (value & 2) << 1 | (value & 1)
    else:
        raise ValueError, "Only depths of 1, 2 and 4 are supported, not %d." % depth

    return result

def get_ula():

    "Return a ULA initialised with a memory array and video."

    return ULA(get_cpu(), get_ram(), get_video())

def get_video():

    "Return a video circuit."

    return Video()

def get_ram():

    "Return an instance representing the computer's RAM hardware."

    return RAM()

def get_cpu():

    "Return an instance representing the CPU."

    return CPU()

# Test program providing coverage (necessary for compilers like Shedskin).

if __name__ == "__main__":
    ula = get_ula()
    ula.set_mode(2)
    ula.reset()
    ula.ram.fill(0x5800 - 320, 0x8000, encode((2, 7), 4))

    # Make a simple two-dimensional array of tuples (three-dimensional in pygame
    # terminology).

    a = update(ula)

# vim: tabstop=4 expandtab shiftwidth=4
